# This file contains the basic functions used by find_exceptional_fillings.py

from imath import isqrt
import hikmot.interval as itl

# Eliminates duplicates from a list
# (Works only if the list is already ordered, so that duplicates are nearby)

def eliminate_duplicates(list):
	l = len(list)
	i = 0
	while i<l-1:
		if list[i] == list[i+1]:
			del list[i]
			l = l-1
		else:
			i = i+1

# Returns True iff the pair of integers (a,b) determines a slope 
# ie if they are coprime and b>=0 and different from (-1,0)

def real_slope(a, b):
	if (a==0 and b<>1) or (b==0 and a<>1): return False
	if b<0: return False
	i = 2
	if a<0: a=-a
	while i <= a and i <= b:
		if a%i == 0 and b%i == 0: return False
		i = i+1
	return True

# Tells whether 'filling' contains 'basic_filling', that is if 'filling'
# may be obtained from 'basic_filling' by filling some more slopes. 
# For instance, the filling [(1, 0), (-1, 1), (0, 0)] 
# contains the filling      [(0, 0), (-1, 1), (0, 0)]
# (if filling == basic_fillings it returns True)

def contains(filling, basic_filling):
	l = len(filling)
	for j in range(l):
		if (filling[j] != basic_filling[j]) and (basic_filling[j] != (0,0)):
			return False
	return True

# Tells whether 'filling' contains an element in the list 'basic_fillings'. 
# 'Contains' here has the meaning explained in the previous routine!
# For instance, the filling [(1, 1), (2,1)] contains the second element of the list
# [[(1, 0), (0, 0)], [(1, 1), (0, 0)]]

def list_contains(filling, basic_fillings):
	for basic_filling in basic_fillings:
		if contains(filling, basic_filling):
			return True
	return False

# Tells whether 'list' contains 'filling'. 
# 'Contains' here has the meaning explained in the previous routine!
# For instance, the filling [(1, 1), (2,1)] is contained in the second element of the list
# [[(1, 0), (0, 0)], [(1, 1), (0, 0)]]
# Remember that 'list' is not really a list, but is had a kind of lexicographic tree structure

def list_contains_new(filling, list):
	if not filling:
		return True
	for node in list:
		if node[0] == filling[0] or node[0] == (0,0):
			new_filling = filling[1:]
			new_list = node[1]
			if list_contains_new(new_filling, new_list):
				return True
	return False

# Returns the number of fillings in a list
# Remember that 'list' is not really a list, but is had a kind of lexicographic tree structure
# NOTE: Returns 1 if it is empty! Returns the correct number otherwise...

def num_fillings(list):
	if len(list) == 0: return 1
	total = 0
	for node in list:
		total = total + num_fillings(node[1])
	return total

# Adds a filling to a list of filling. 
# To speed up operations, the list has a kind of lexicographic tree structure.

def add_filling_to_list(filling, list):
	list_node = list
	for filling_slope in filling:
		slope_is_already_there = False
		for node in list_node:
			if node[0] == filling_slope:
				slope_is_already_there = True
				list_node = node[1]
				break
		if slope_is_already_there == False:
			new_node = [filling_slope, []] 	#va bene?
			list_node.append(new_node)	
			list_node = new_node[1]		#va bene?

# Converts the type of list that we use here to an ordinary list of fillings

def our_list_to_ordinary_list(list):
	answer = []
	for node in list:
		if not node[1]:
			answer.append([node[0]])
		else:
			sub_answer = our_list_to_ordinary_list(node[1])
			for i in sub_answer:
				i2 = i[:]
				i2.insert(0, node[0])
				answer.append(i2)
	return answer		

# Adds all the fillings of list2 to list1
# We lazily use other routines, probably this is not very efficient.

def append_list_to_list(list1, list2): 
	ordinary_list2 = our_list_to_ordinary_list(list2)
	for filling in ordinary_list2:
		add_filling_to_list(filling, list1)
				
# Returns the (squared) lenght of a geodesic determined py the pair (p, q) in an euclidean torus
# with area A and shape (x + iy)

def squared_length(A, x, y, p, q):

	# The formula is (A/y) * ((p+x*q)**2 + (y*q)**2)
	a = (p + x * q)

	return (A/y) * (a * a + ((y * q) * (y * q)))

# Returns the list of all short slopes (i.e. those having length <= K = 6.1) 
# in a euclidean torus with area A and shape x + iy. 

def get_short_slopes(A, x, y):

	K = itl.interval(6.0001)
	short_slopes = []

	# If l^2(px+qy)<=K^2, then (A/y)*(yq)^2 <= K^2 which implies that q^2 <= K^2/(Ay)
	# Then qmax = K/sqrt(A*y)

	qmax = K/isqrt(A*y)   
	for q in range (0, int(qmax.sup)+1):

		# We have (p+xq)^2 + (yq)^2 <= K^2y/A and therefore (p+xq^2) <= H^2 with
		# H = sqrt((K*K*y)/A - (y*q)**2)

		H = isqrt((K * K * y)/A - (y * q) * (y * q))

		# p ranges from -H - x*q to H - x*q

		pmin = -H - x * q
		pmax = H - x * q

		# We now select among finitely many slopes (p, q) the really short ones

		for p in range (int(pmin.inf)-1, int(pmax.sup)+1):
			if real_slope(p,q)==1:

				# checks that it is indeed a short slope:

				l = squared_length(A, x, y, p, q)
				if l.inf < (K * K).sup:
					slope = (p,q)
					short_slopes.append(slope)
	return short_slopes

# Transforms (p, q) into the string 'p' if q=1 and '\tfrac {p}q' if not.
# If (0,0) returns '*'

def slope_to_latex(slope):
	if slope == (0,0): 
		return '*'
	if slope == (1,0):
		return '\infty'
	if slope[1] == 1:
		return str(slope[0])
	stringa = ''
	if slope[0] <0:
		stringa = stringa + '-'
	stringa = stringa + r'\tfrac{' + str(abs(slope[0])) + '}{' + str(abs(slope[1])) + '}'
	return stringa
	
def filling_to_latex(filling):
	stringa = '$('
	for slope in filling:
		stringa = stringa + slope_to_latex(slope) + ','
	stringa = stringa[:-1] + ')$'
	return stringa
	
def list_to_latex(list_of_fillings):
	passo = 4
	stringa = ''
	counter = 0
	for filling in list_of_fillings:
		stringa = stringa + filling_to_latex(filling)
		counter = counter + 1
		if counter % 4 == 0:
			stringa = stringa + r' \\ '
		else:
			stringa = stringa + ' & '
	return stringa