# This file contains some functions that use the manifold class.

from constants import *

# functions.py contains the basic numerical functions used here

from functions import *

import hikmot


# A dictionary gives a number for each degeneracy level of solution

degeneracy_level = {'all tetrahedra positively oriented': 0,
'contains negatively oriented tetrahedra': 1,
'contains flat tetrahedra': 2,
'contains degenerate tetrahedra': 3,
'unrecognized solution type': 4,
'no solution found': 5,
'not attempted': 6}

# Returns the minimum imaginary part of all tetrahedra shapes

def min_imaginary_shapes(M):
	imaginary_shapes = []
	for i in M.tetrahedra_shapes(part = 'rect'):
		imaginary_shapes.append(i.imag)
	return min(imaginary_shapes)

# Checks that the solution is positive
# Here M.solution_type == 'all tetrahedra positively oriented' is not enough,
# we require that the imaginary part of every tetrahedron shape 
# is bigger than the constant 'min_imaginary'

def positive_solution(M, min_imaginary):
	return min_imaginary_shapes(M) > min_imaginary

# Returns True if the filled manifold is closed

def is_closed(M):
	num_cusps = M.num_cusps() 
	for i in range(num_cusps):
		if M.cusp_info('complete?')[i] == True:
			return False
	return True	

# Returns the number of cusps that are filled

def num_filled_cusps(M):
	num_cusps = M.num_cusps() 
	counter = 0
	for i in range(num_cusps):
		if M.cusp_info('complete?')[i] == False:
			counter = counter + 1
	return counter

# Tries to hyperbolize the manifold M. 
# It returns a copy of M whose solution has the least degree of degeneracy 
# among those investigated.
# More precisely, it returns a triple flag, manifold, tetrahedra_shapes
# where flag == True if 'manifold' is verified hyperbolic
#
# first_to_be_filled_permanently is useful when the function calls itself, to avoid
# analyzing multiple times the same fillings.

def try_to_hyperbolize(M, number_of_randomizes, try_permanent_fillings, min_imaginary, first_to_be_filled_permanently = 0):
	answer = M.copy()

	# I tried to add 'answer.simplify()' here but things get typically worse
	
	# The best solution is: 
	# 1) 'all tetrahedra positively oriented'
	# 2) positive solution == True, which checks that imaginary parts are big enough  
	# 3) hikmot verified
	# If the algorithm finds such a best solution, it stops and returns it.

	if degeneracy_level[answer.solution_type()] == 0 and positive_solution(answer, min_imaginary):
		flag, tetrahedra_shapes = hikmot.verify_hyperbolicity(answer, verbose >=2)
		if flag:
			return True, answer, tetrahedra_shapes
		else:
			print 'no flag'
	Mcopy = answer.copy()

	# Try to randomize the triangulation 'number_of_randomizes' times.

	for tentative in range(number_of_randomizes):
		Mcopy.randomize()
		if degeneracy_level[Mcopy.solution_type()] == 0 and positive_solution(Mcopy, min_imaginary):
			flag, tetrahedra_shapes = hikmot.verify_hyperbolicity(Mcopy, verbose >=2)
			if flag:
				return True, Mcopy, tetrahedra_shapes
		if degeneracy_level[Mcopy.solution_type()] < degeneracy_level[answer.solution_type()]:
			answer = Mcopy.copy()

	# Try to permanently fill some cusps (if try_permanent_fillings == True):
	# We impose counter >= first_to_be_filled_permanently to ensure that
	# the to-be-permanently-filled subsets of {filled cusps} are not counted multiple times

	if try_permanent_fillings == True: 
		num_cusps = M.num_cusps() 
		if num_cusps >= 2:		
			if try_only_maximal_permanent_fillings == False:
				counter = 0
				for i in range(num_cusps):
					if M.cusp_info('complete?')[i] == False:
						if counter >= first_to_be_filled_permanently:
							Mfilled = M.filled_triangulation([i])	
							flag, another_answer, other_shapes = try_to_hyperbolize(Mfilled, number_of_randomizes, try_permanent_fillings, min_imaginary, counter)
							if flag == True:
								return True, another_answer, other_shapes
							if degeneracy_level[another_answer.solution_type()] < degeneracy_level[answer.solution_type()]:
								answer = another_answer.copy()
						counter = counter + 1
			else:
				if is_closed(M) == True:
					for i in range(num_cusps):
						to_be_filled = range(num_cusps)
						to_be_filled.remove(i)
						Mfilled = M.filled_triangulation(to_be_filled)
						flag, another_answer, other_shapes = try_to_hyperbolize(Mfilled, number_of_randomizes, try_permanent_fillings, min_imaginary)
						if flag == True:
							return True, another_answer, other_shapes
						if degeneracy_level[another_answer.solution_type()] < degeneracy_level[answer.solution_type()]:
							answer = another_answer.copy()										
				else:
					if num_filled_cusps(M) > 0:
						Mfilled = M.filled_triangulation()
						flag, another_answer, other_shapes = try_to_hyperbolize(Mfilled, number_of_randomizes, try_permanent_fillings, min_imaginary)
						if flag == True:
							return True, another_answer, other_shapes
						if degeneracy_level[another_answer.solution_type()] < degeneracy_level[answer.solution_type()]:
							answer = another_answer.copy()
			
			
	return False, answer, []

# Applies an isometry to a filling

def apply_isometry(isometry, filling):
	num_cusps = len(filling)
	new_filling = filling[:]
	for i in range(num_cusps):
		matrix = isometry.cusp_maps()[i].data
		slope = filling[i]
		mapped_slope = (matrix[0][0] * slope[0] + matrix[0][1] * slope[1], matrix[1][0] * slope[0] + matrix[1][1] * slope [1])

		# Recall that a slope here is a pair (p,q) with q>=0 and is not (-1,0):

		if mapped_slope[1]<0:
			mapped_slope = (-mapped_slope[0], -mapped_slope[1])
		if mapped_slope == (-1,0):
			mapped_slope = (1,0)
		new_filling[isometry.cusp_images()[i]] = mapped_slope
	return new_filling

# Quotients the list 'fillings' by the action of the elements in 'isometries'
# Returns by default an ordered list

def quotient_list_by_isometries(fillings, isometries):
	answer = []
	for filling in fillings:

		# Among all images of filling through isometries, we take the lexicographically smallest one:
		# This gives a unique representative on each equivalence class of fillings.

		new_filling = filling[:]
		for isometry in isometries:
			image_filling = apply_isometry(isometry, filling)
			if image_filling < new_filling:
				new_filling = image_filling 
		answer.append(new_filling[:])
	answer.sort()
	eliminate_duplicates(answer)
	return answer
	
# Returns the orbit of a filling under the action of isometries
# Returns by default an ordered list without duplicates
	
def get_orbit(filling, isometries):
	answer = []
	for isometry in isometries:
		image_filling = apply_isometry(isometry, filling)
		answer.append(image_filling[:])
	answer.sort()
	eliminate_duplicates(answer)
	return answer

# Returns the union of all the orbits of all the fillings in 'fillings' 
# Returns by default an ordered list without duplicates

def expand_list_by_isometries(fillings, isometries):
	answer = []
	for filling in fillings:
		orbit = get_orbit(filling, isometries)
		answer = answer + orbit[:]
	answer.sort()
	eliminate_duplicates(answer)
	return answer

